/*!
 * (C) Copyright IBM Corp. 2005, 2005
 *
 * \file kboot.c
 * \brief main function for kboot boot loader
 *
 * $Id: kboot.c,v 1.7 2006/02/10 15:09:35 loehr Exp $
 *
 * Author(s): Ralph Wuerthner (rwuerthn@de.ibm.com)
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "kboot.h"


char *arg0; //<! global variable pointing to argv[0] (used in MEM_ASSERT)


/**
 * Print help message
 *
 * \param[in] arg0 \p argv[0] from \p main()
 */

void
print_help(const char *arg0)
{
	printf("Usage: %s [OPTION] <configuration file URI>\n"
	    "Start kboot user interface.\n\n"
	    " -o <device>   Output kboot status messages to <device>. "
	    "Default is stdout.\n"
	    " -v            Output version information and exit.\n"
	    " -h            Display this help and exit.\n",
	    arg0);
}


/**
 * Parse arguments passed on the command line to kboot and store parsed
 * arguments in \p kboot_args. In case of command line errors exit program
 * with return code 1.
 *
 * \param[out] kboot_args  Pointer to \p kboot_arguments structure with parsed
 *  command line arguments.
 * \param[in]  argc   Number of arguments passed to program.
 * \param[in]  argv   Array with arguments passed to program.
 */

enum parse_result
parse_arguments(struct kboot_arguments *kboot_args, int argc, char **argv)
{
	char *optstring="o:vh";
	int c;

	// initialize arguments
	memset(kboot_args, 0x0, sizeof(*kboot_args));
	cfg_strinit(&kboot_args->config_uri);
	cfg_strinit(&kboot_args->console);

	opterr=1;
	optind=1;
	while ((c = getopt(argc, argv, optstring)) != -1) {
		switch (c) {
		case 'o':
			cfg_strcpy(&kboot_args->console, optarg);
			break;

		case 'v':
			printf("kboot user interface version %s\n"
			    "Written by Ralph Wuerthner and Michael Loehr.\n",
			    KBOOT_UI_VERSION);
			return PA_DONE;

		case 'h':
			print_help(argv[0]);
			return PA_DONE;

		case ':':
			fprintf(stderr, "%s: option requires an argument"
			    "-- %c\n", argv[0], optopt);
			fprintf(stderr, "Try `%s -h' for more information.\n",
			    argv[0]);
			return PA_ERROR;
		}
	}

	if (optind != argc-1) {
		fprintf(stderr, "%s: invalid argument(s). "
		    "Try `%s -h' for more information.\n",
		    argv[0], argv[0]);
		return PA_ERROR;
	}

	cfg_strcpy(&kboot_args->config_uri, argv[optind]);
	return PA_VALID;
}


/**
 * Redirect stdout & stderr to different device if requested on command line.
 *
 * \param[in]  device Path to output device.
 */

void
redirect_output(const char *device)
{
	int fd;
	struct stat st;

	if (strlen(device)) {
		if (stat(device, &st)) {
			printf("Warning: unable to check output device '%s'"
			    " - %s\n", device, strerror(errno));
			goto error;
		}
		if ((st.st_mode & S_IFMT) != S_IFCHR) {
			printf("Warning: '%s' is not a character device\n",
			    device);
			goto error;
		}
		fd = open(device, O_RDWR | O_NONBLOCK, 0);
		if (fd != -1) {
			dup2(fd, 1);
			dup2(fd, 2);
			close(fd);
			return;
		}
		printf("Warning: error opening output device '%s' - %s\n",
		    device, strerror(errno));

	error:
		printf("Continuing using stdout as output device.\n");
	}
}


/**
 * \p main -- the entry point into a world of wonderful possibilities.
 *
 * \param[in]  argc Number of arguments passed to program.
 * \param[in]  argv Array with arguments passed to program.
 */

int
main(int argc, char **argv)
{
	char *errmsg, *startup_msg;
	struct kboot_arguments kboot_args;
	struct cfg_toplevel config;
	struct cfg_bentry boot;
	enum parse_result pa_return;
	int retval = 0; 

	arg0 = argv[0];

	pa_return = parse_arguments(&kboot_args, argc, argv);

	switch (pa_return) {

	case PA_DONE:
		retval = 0;
		break;

	case PA_ERROR:
		retval = 1;
		break;

	case PA_VALID:
		redirect_output(kboot_args.console);
		cfg_strinit(&startup_msg);

		// print startup message to redirected console
		printf("kboot user interface is starting.\n");
		printf("Configuration file source: %s\n",
		    kboot_args.config_uri);

		// parse configuration file
		errmsg = parse_kboot(&config, kboot_args.config_uri);
		if (errmsg) {
			cfg_strprintf(&startup_msg,
			    "Error loading configuration data:\n"
			    "%s\n"
			    "Continuing with blank configuration\n", errmsg);
			free(errmsg);
			cfg_init(&config);
		}

		while (WORLD_EXISTS) {
			// launch user interface modules
			if (userinterface(startup_msg, &config, &boot)) {
				printf("Error: unable to find boot configuration\n"
				    "kboot is exiting - goodbye!\n");
				retval = 1;
				break;
			}
			cfg_strcpy(&startup_msg, "");

			// start new kernel
			errmsg = loader(&boot);
			if (errmsg) {
				cfg_strprintf(&startup_msg,
				    "Unable to  start selected configuration:\n"
				    "%s\n"
				    "Please try again.\n",
				    errmsg);
				free(errmsg);
			} else
				cfg_strprintf(&startup_msg,
				    "Error starting selected configuration. "
				    "Please try again.\n");
			cfg_bentry_destroy(&boot);
		}
		break;
	}

	// always reached

	return retval;
}
