/*!
 * (C) Copyright IBM Corp. 2006,2006
 *
 * \file parser.c
 * \brief General functions for parsing config files.
 *
 * $Id: parser.c,v 1.5 2006/02/10 15:09:35 loehr Exp $
 *
 * Author(s): Ralph Wuerthner (rwuerthn@de.ibm.com)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "kboot.h"


#define INITIAL_LINE_BUF 2 //!< initial size of buffer for reading lines


/**
 * Initialize a parser_context structure.
 *
 * \param[in] context Pointer to parser_context structure to be initialized.
 */

void
parser_init(struct parser_context *context)
{
	memset(context, 0x0, sizeof(*context));

	// initialize string members
	cfg_strinit(&context->errmsg);
	cfg_strinitcpy(&context->boot_default, "0");

	// initialize read line buffer
	context->line = malloc(INITIAL_LINE_BUF);
	MEM_ASSERT(context->line);
	context->line_size = INITIAL_LINE_BUF;
}


/**
 * Destroy a parser_context structure and free all alocated memory.
 *
 * \param[in] context Pointer to parser_context structure to be destroyed.
 */

void
parser_destroy(struct parser_context *context)
{
	int n, depth = context->depth;

	// close all open files & free read line buffer
	for (n = 0; n < depth; n++)
		parser_close_incl(context);
	free(context->line);
	context->line = NULL;
	context->depth = 0;

	// destroy string members
	cfg_strfree(&context->errmsg);
	cfg_strfree(&context->boot_default);
}


/**
 * Set error message string in parser context.
 *
 * \param[in] context Pointer to parser_context structure
 * \param[in] str Error message string to be set in parser context
 */

void parser_set_errmsg(struct parser_context *context, const char *str)
{
	cfg_strcpy(&context->errmsg, str);
}


/**
 * Set error message string in parser context using a printf style interface.
 *
 * \param[in] context  Pointer to parser_context structure.
 * \param[in] format   printf format string.
 * \param[in] ...      Additional printf arguments.
 */

void parser_printf_errmsg(struct parser_context *context,
    const char *format, ...)
{
	va_list arg;

	va_start(arg, format);
	cfg_strvprintf(&context->errmsg, format, arg);
}


/**
 * Initialize all data fields in parser include file structure.
 *
 * \param[in] incl   Pointer to incl_stack structure to be initialized.
 */

void
parser_incl_init(struct incl_stack *incl)
{
	incl->mode = PARSE_URI;
	incl->file = NULL;
	cfg_strinit(&incl->uri);
	cfg_strinit(&incl->filename);
	incl->strsource = NULL;
	incl->strreadpos = NULL;
	incl->lineno = 0;
}


/**
 * Open include file. Any subsequent calls to \p parser_getstr will
 * read data from this URI.
 *
 * \param[in] context Pointer to parser_context structure.
 * \param[in] uri     String with URI to be opened.
 * \return    \p P_OK on success, \p P_ERR on error.
 */

enum p_return
parser_open_incl_uri(struct parser_context *context, const char *uri)
{
	char *errmsg;
	struct incl_stack *incl = &(context->incl[context->depth]);

	if (context->depth >= MAX_INCL_DEPTH) {
		parser_set_errmsg(context,
		    "maximum include file depth reached");
		return P_ERR;
	}
	parser_incl_init(incl);
	incl->mode = PARSE_URI;

	// create local copy of URI
	cfg_strprintf(&incl->filename, URI_TEMP_FILENAME, context->depth);
	if (comp_load(incl->filename, uri, NULL, &errmsg)) {
		if (errmsg) {
			parser_printf_errmsg(context,
			    "unable to open URI - %s", errmsg);
			free(errmsg);
		} else
			parser_set_errmsg(context, "error while opening URI");
		return P_ERR;
	}

	// open local copy
	incl->file = fopen(incl->filename, "r");
	if (incl->file == NULL) {
		parser_printf_errmsg(context,
		    "unable to open local URI copy '%s' - %s",
		    incl->filename, strerror(errno));
		return P_ERR;
	}
	cfg_strcpy(&incl->uri, uri);
	context->depth++;

	return P_OK;
}


/**
 * Open include string. Any subsequent calls to \p parser_getstr will
 * read data from this string.
 *
 * \param[in] context Pointer to parser_context structure.
 * \param[in] str     String with configuration data to read from
 * \return    \p P_OK on success, \p P_ERR on error.
 */

enum p_return
parser_open_incl_string(struct parser_context *context, const char *str)
{
	struct incl_stack *incl = &(context->incl[context->depth]);

	if (context->depth >= MAX_INCL_DEPTH) {
		parser_set_errmsg(context,
		    "maximum include file depth reached");
		return P_ERR;
	}
	parser_incl_init(incl);
	incl->mode = PARSE_STRING;

	incl->strsource = str;
	incl->strreadpos = str;
	cfg_strcpy(&incl->uri, "string");
	context->depth++;

	return P_OK;
}


/**
 * Close include file. Any subsequent calls to \p parser_getstr will
 * read data from the previously opened data source.
 *
 * \param[in] context Pointer to parser_context structure.
 * \return    \p P_OK on success, \p P_ERR on error.
 */

enum p_return
parser_close_incl(struct parser_context *context)
{
	struct incl_stack *incl = &(context->incl[context->depth-1]);

	if (context->depth <= 0) {
		parser_set_errmsg(context, "no more include files are open");
		return P_ERR;
	}
	switch (incl->mode) {
	case PARSE_URI:
		fclose(incl->file);
		unlink(incl->filename);
		break;

	case PARSE_STRING:
		break;
	}

	cfg_strfree(&incl->uri);
	cfg_strfree(&incl->filename);
	incl->strsource = NULL;
	incl->strreadpos = NULL;
	incl->file = NULL;
	incl->lineno = 0;
	context->depth--;

	return P_OK;
}


/**
 * Check current include file for end of file.
 *
 * \param[in] context Pointer to parser_context structure.
 * \return    \p P_OK if end of file is not reached,
 *            \p P_EOF if end of file is reached, \p P_ERR on error.
 */

enum p_return
parser_eof(struct parser_context *context)
{
	struct incl_stack *incl = &(context->incl[context->depth-1]);

	if (context->depth <= 0) {
		parser_set_errmsg(context, "no file open");
		return P_ERR;
	}

	switch (incl->mode) {
	case PARSE_URI:
		if (feof(incl->file))
			return P_EOF;
		else
			return P_OK;

	case PARSE_STRING:
		if (strlen(incl->strreadpos))
			return P_OK;
		else
			return P_EOF;
	}

	return P_OK;
}


/**
 * Read at most \p size characters from current data source and store
 * data in \p str. On success \p P_OK is returned. If EOF is reached without
 * reading any data \p P_EOF is returned. On error an error message
 * is set and \p P_ERR is returned.
 *
 * \param[in] context Pointer to parser_context structure.
 * \param[in] str     Pointer to string buffer.
 * \param[in] size    Size of string buffer.
 * \return    \p P_OK on success, \p P_EOF if EOF is reached without
 *            reading any data, \p P_ERR on error.
 */

enum p_return
parser_getstr(struct parser_context *context, char *str, int size)
{
	char *ptr;
	size_t len;
	struct incl_stack *incl = &(context->incl[context->depth-1]);

	switch (incl->mode) {
	case PARSE_URI:
		ptr = fgets(str, size, incl->file);
		if (ptr == NULL) {
			if (errno) {
				parser_printf_errmsg(context,
				    "error reading data - %s",
				    strerror(errno));
				return P_ERR;
			} else
				return P_EOF;
		}
		return P_OK;
		break;

	case PARSE_STRING:
		ptr = strchr(incl->strreadpos, '\n');
		if (ptr)
			len = ptr-incl->strreadpos+1;
		else
			len = strlen(incl->strreadpos);

		if (len>0) {
			if (len>size-1)
				len = size-1;
			memcpy(str, incl->strreadpos, len);
			str[len] = '\0';
			incl->strreadpos += len;
			return P_OK;
		}
		else
			return P_EOF;
	}

	return P_EOF;
}


/**
 * Read next line from current include file and copy data to
 * \p line field in parser_context structure.
 *
 * \param[in] context Pointer to parser_context structure.
 * \return    \p P_OK on success, \p P_EOF if end of file is reached,
 *            \p P_ERR on error.
 */

enum p_return
parser_readln(struct parser_context *context)
{
	char *ptr = context->line;
	char *nl_ptr;
	enum p_return pret;

	context->line[0]='\0';
	if (context->depth <= 0) {
		parser_set_errmsg(context, "no file open");

		return P_ERR;
	}

	// read next line from current input file
	// and expand line buffer if necessary
	do {
		pret = parser_getstr(context, ptr,
		    context->line_size-(ptr-context->line));
		if (pret == P_ERR)
			return P_ERR;
		if (pret == P_EOF) {
			if (strlen(context->line))
				break;
			else
				return P_EOF;
		}

		nl_ptr = strchr(context->line, '\n');
		if (nl_ptr == NULL) {
			context->line_size *= 2;
			context->line = realloc(context->line,
			    context->line_size);
			MEM_ASSERT(context->line);
		}
		ptr = context->line+strlen(context->line);
	} while (nl_ptr == NULL);

	// remove trailing '\n'
	nl_ptr = strchr(context->line, '\n');
	if (nl_ptr)
		*nl_ptr='\0';
	context->incl[context->depth-1].lineno++;

	return P_OK;
}


/**
 * Read next line from current data source. If EOF is reached close
 * current data source and continue reading on previously opened
 * data source. If no more open data sources are available \p P_EOF
 * is returned. On error \p P_ERR is returned.
 *
 * \param[in] context Pointer to parser_context structure.
 * \return    \p P_OK on success, \p P_EOF if no more open data sources
 *             are available, \p P_ERR on error.
 */

enum p_return
parser_readln_incl(struct parser_context *context)
{
	enum p_return pret;

	do {
		pret = parser_readln(context);
		if (pret != P_EOF)
			return pret;

		parser_close_incl(context);
	} while (context->depth>0);

	return P_EOF;
}


/**
 * Read next line from current include file but ignore empty & comment
 * lines and copy data to \p line element in parser_context structure.
 *
 * \param[in] context Pointer to parser_context structure.
 * \return    \p P_OK on success, \p P_EOF if no more open data sources
 *             are available, \p P_ERR on error.
 */

enum p_return
parser_readln_notempty(struct parser_context *context)
{
	enum p_return pret;
	char *line;
	size_t spcs;

	do {
		pret = parser_readln_incl(context);
		if (pret)
			return pret;
		line = context->line;
		spcs = strspn(line, P_WHITESPACE);
	} while (line[spcs] == P_COMMENT || strlen(line) == spcs);

	return P_OK;
}


/**
 * Extract first token from \p str and compare against token array. If token
 * was found in array return array index, with numbering starting with \p 0.
 * If token wasn't found \p -1 is returned.
 *
 * \param[in]  str      Pointer to string to be parsed.
 * \param[in]  tokens   Pointer to string array with tokens. End of token
 *                      array is marked by \p NULL.
 * \param[out] next     Pointer to first character after token.
 * \return     Token index, with numbering starting with \p 0, \p -1 if token
 *             wasn't found in token array.
 */

int
parser_find_token(const char *str, const char **tokens, const char **next)
{
	const char *ptr;
	int idx = 0;
	size_t len;

	ptr = str+strspn(str, P_WHITESPACE);
	len = strcspn(ptr, P_WHITESPACE);

	while (*tokens != NULL) {
		if (len == strlen(*tokens) &&
		    strncmp(ptr, *tokens, len) == 0) {
			*next = ptr+len;
			return idx;
		}
		tokens++;
		idx++;
	}

	return -1;
}


/**
 * Return string pointer to first none white space character.
 *
 * \param[in] str Pointer to string.
 * \return    Pointer to first none white space character in input string.
 */

const char *
parser_skip_whitespace(const char *str)
{
	return str+strspn(str, P_WHITESPACE);
}


/**
 * Return pointer to first argument in \p str by skipping leading white
 * space characters and calculate length of first argument.
 *
 * \param[in]   str  Pointer to string.
 * \param[out]  len  Length of argument.
 * \return      Pointer to first argument character
 */

const char *
parser_next_arg(const char *str, size_t *len)
{
	str = parser_skip_whitespace(str);
	*len = strcspn(str, P_WHITESPACE);
	return str;
}
