/*!
 * (C) Copyright IBM Corp. 2005, 2005
 *
 * \file parser.h
 * \brief Include file for kboot config file parser.
 *
 * $Id: parser.h,v 1.5 2006/02/09 16:17:57 loehr Exp $
 *
 * Author(s): Ralph Wuerthner (rwuerthn@de.ibm.com)
 *
 */


#ifndef _PARSER_H_
#define _PARSER_H_

#include "config.h"

#define MAX_INCL_DEPTH 20  //!< max. include file depth
#define P_COMMENT '#'       //!< comment character in config files
#define P_WHITESPACE " \t"  //!< string with white space characters
#define URI_TEMP_FILENAME "/tmp/kbootparser-%d.cfg"
//!< filename for local copies of URIs



/**
 * Supported operational modes for parser
 */

enum parser_mode {
  PARSE_KBOOT_CONFIGURATION,  //!< parse full kboot configuration file
  PARSE_KBOOT_BOOT_ENTRY,     //!< parse kboot boot entry only
};


/**
 * Return codes for parser functions.
 */

enum p_return {
  P_ERR = -1,  //!< parsing error detected
  P_OK = 0,    //!< success
  P_EOF = 1,   //!< end of file reached
};


/**
 * Read mode for \p parser_fgets() function
 */

enum parser_source {
  PARSE_URI,    //!< data source is a URI
  PARSE_STRING, //!< data source is a string
};


/**
 *
 * This structure defines data required for the include file stack.
 */

struct incl_stack {
  enum parser_source mode; //!< read mode
  FILE *file;         //!< stream for local URI copy
  char *uri;          //!< original URI used to access file
  char *filename;     //!< filname used for local URI copy
  const char *strsource; //!< source string
  const char *strreadpos; //!< read position in source string
  int lineno;         //!< current read line number
};


/**
 *
 * This structure defines the context reqired by the parser to work.
 */

struct parser_context {
  enum parser_mode mode;     //!< parser operational mode
  char *line;                //!< pointer to current line
  int line_size;             //!< size of buffer allocated for current line
  struct incl_stack incl[MAX_INCL_DEPTH]; //!< include file stack array
  int depth;                 //!< current include depth
  char *errmsg;              //!< buffer for parser error messages
  char *boot_default;        //!< buffer for default boot label during parsing
};

void parser_init(struct parser_context *context);
void parser_destroy(struct parser_context *context);
void parser_set_errmsg(struct parser_context *context,const char *str);
void parser_printf_errmsg(struct parser_context *context,
			  const char *format,...);
void parser_incl_init(struct incl_stack *incl);
enum p_return parser_open_incl_uri(struct parser_context *context,
			       const char *filename);
enum p_return parser_open_incl_string(struct parser_context *context,
			       const char *str);
enum p_return parser_close_incl(struct parser_context *context);
enum p_return parser_eof(struct parser_context *context);
enum p_return parser_getstr(struct parser_context *context, char *str,
			    int size);
enum p_return parser_readln(struct parser_context *context);
enum p_return parser_readln_incl(struct parser_context *context);
enum p_return parser_readln_notempty(struct parser_context *context);
int parser_find_token(const char *str, const char **tokens,
		      const char **next);
const char *parser_skip_whitespace(const char *str);
const char *parser_next_arg(const char *, size_t *);
char *parse_kboot(struct cfg_toplevel *config, const char *uri);
char *parse_kboot_boot_entry(struct cfg_bentry *config, const char *str);

#endif /* #ifndef _PARSER_H_ */
