/*!
 * (C) Copyright IBM Corp. 2005, 2005
 *
 * \file parser_kboot.c
 * \brief Parser for kboot config files.
 *
 * $Id: parser_kboot.c,v 1.7 2006/02/09 16:17:57 loehr Exp $
 *
 * Author(s): Ralph Wuerthner (rwuerthn@de.ibm.com)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "parser.h"


// command tokens
static const char *config_cmds[] = {
	"default",
	"timeout",
	"password",
	"userinterface",
	"include",
	"boot_entry",
	"title",
	"label",
	"root",
	"kernel",
	"initrd",
	"cmdline",
	"parmfile",
	"lock",
	"pause",
	"halt",
	"reboot",
	"}",
	NULL
};

enum config_token {
	T_DEFAULT,
	T_TIMEOUT,
	T_PASSWORD,
	T_USERINTERFACE,
	T_INCLUDE,
	T_BOOT_ENTRY,
	T_TITLE,
	T_LABEL,
	T_ROOT,
	T_KERNEL,
	T_INITRD,
	T_CMDLINE,
	T_PARMFILE,
	T_LOCK,
	T_PAUSE,
	T_HALT,
	T_REBOOT,
	T_CBRACE,
};


/**
 * Parse \p default command and update configuration structure.
 *
 * \param[in,out] config    Pointer to cfg_toplevel structure.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_default(struct cfg_toplevel *config, struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;

	arg_ptr = parser_next_arg(next, &arg_len);

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(arg_ptr+arg_len) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	cfg_strncpy(&context->boot_default, arg_ptr, arg_len);

	return P_OK;
}


/**
 * Parse \p timeout command and update configuration structure.
 *
 * \param[in,out] config    Pointer to cfg_toplevel structure.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_timeout(struct cfg_toplevel *config, struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	int timeout;
	size_t arg_len;

	arg_ptr = parser_next_arg(next, &arg_len);

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(arg_ptr+arg_len) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	if (sscanf(arg_ptr, "%d", &timeout) != 1 || timeout < 0 ) {
		parser_set_errmsg(context, "invalid argument");
		return P_ERR;
	}
	config->timeout = timeout;

	return P_OK;
}


/**
 * Parse \p password command and update configuration structure.
 *
 * \param[in,out] config    Pointer to cfg_toplevel structure.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_password(struct cfg_toplevel *config,
    struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;

	arg_ptr = parser_next_arg(next, &arg_len);

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(arg_ptr+arg_len) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	cfg_strncpy(&config->password, arg_ptr, arg_len);

	return P_OK;
}


/**
 * Parse \p userinterface command and update configuration structure.
 *
 * \param[in,out] config    Pointer to cfg_toplevel structure.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_userinterface(struct cfg_toplevel *config,
    struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;
	struct cfg_userinterface ui;

	// there must be least one argument
	arg_ptr = parser_next_arg(next, &arg_len);
	if (*arg_ptr == '\0') {
		parser_set_errmsg(context, "missing argument");
		return P_ERR;
	}

	cfg_userinterface_init(&ui);
	cfg_strncpy(&ui.module, arg_ptr, arg_len);

	// find module arguments
	arg_ptr = parser_skip_whitespace(arg_ptr+arg_len);
	arg_len = strlen(arg_ptr);
	for (; arg_len > 0 && strchr(P_WHITESPACE,
		*(arg_ptr+arg_len-1)) != NULL; arg_len--);
	if (arg_len)
		cfg_strncpy(&ui.cmdline, arg_ptr, arg_len);

	cfg_add_userinterface(config, &ui);
	cfg_userinterface_destroy(&ui);

	return P_OK;
}


/**
 * Parse \p include command and call \p parser_open_incl_uri() to open
 * include URI.
 *
 * \param[in] context   Pointer to parser_context structure.
 * \param[in] next      Pointer to first character after command token.
 * \return    \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_include(struct parser_context *context, const char *next)
{
	const char *arg_ptr;
	size_t arg_len;
	char *uri;
	enum p_return pret;

	arg_ptr = parser_next_arg(next, &arg_len);

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(arg_ptr+arg_len) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	// open include file
	cfg_strinit(&uri);
	cfg_strncpy(&uri, arg_ptr, arg_len);
	pret = parser_open_incl_uri(context, uri);
	cfg_strfree(&uri);

	return pret;
}


/**
 * Look for open curly brace on current or next configuration file line.
 *
 * \param[in] context   Pointer to parser_context structure.
 * \param[in] next      Pointer to start position for search.
 * \return    \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_next_ocbrace(struct parser_context *context, const char *next)
{
	enum p_return pret;
	const char *arg_ptr;
	size_t arg_len;

	arg_ptr = parser_next_arg(next, &arg_len);
	if (*arg_ptr == '\0') {
		// no more arguments -> '{' must be on new line
		pret = parser_readln_notempty(context);
		if (pret) {
			if (pret == P_EOF)
				parser_set_errmsg(context,
				    "unexpected end of file");
			return P_ERR;
		}

		arg_ptr = parser_next_arg(context->line, &arg_len);
	}

	if (*arg_ptr != '{' || arg_len!=1 ||
	    *parser_skip_whitespace(arg_ptr+arg_len) != '\0' ) {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	return P_OK;
}


/**
 * Parse \p title command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_title(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	const char *arg;
	size_t arg_len;

	// find start of title
	arg = next + strspn(next, P_WHITESPACE);

	// remove trailing whitespaces
	arg_len=strlen(arg);
	for (; arg_len > 0 && strchr(P_WHITESPACE, *(arg+arg_len-1)) != NULL;
	     arg_len--);

	cfg_strncpy(&bentry->title, arg, arg_len);

	return P_OK;
}


/**
 * Parse \p label command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_label(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;

	arg_ptr = parser_next_arg(next, &arg_len);

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(arg_ptr+arg_len) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	cfg_strncpy(&bentry->label, arg_ptr, arg_len);

	return P_OK;
}


/**
 * Parse \p root command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_root(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;

	arg_ptr = parser_next_arg(next, &arg_len);

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(arg_ptr+arg_len) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	cfg_strncpy(&bentry->root, arg_ptr, arg_len);

	return P_OK;
}


/**
 * Parse \p kernel command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_kernel(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;

	if (bentry->action != -1) {
		parser_set_errmsg(context, "invalid boot entry");
		return P_ERR;
	}

	arg_ptr = parser_next_arg(next, &arg_len);

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(arg_ptr+arg_len) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	cfg_strncpy(&bentry->kernel, arg_ptr, arg_len);
	bentry->action = KERNEL_BOOT;

	return P_OK;
}


/**
 * Parse \p initrd command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_initrd(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;

	arg_ptr = parser_next_arg(next, &arg_len);

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(arg_ptr+arg_len) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	cfg_strncpy(&bentry->initrd, arg_ptr, arg_len);

	return P_OK;
}


/**
 * Parse \p cmdline command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_cmdline(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;

	arg_ptr = parser_next_arg(next, &arg_len);
	arg_len = strlen(arg_ptr);
	for (; arg_len > 0 && strchr(P_WHITESPACE,
		*(arg_ptr+arg_len-1)) != NULL; arg_len--);
	if (arg_len)
		cfg_strncpy(&bentry->cmdline, arg_ptr, arg_len);

	return P_OK;
}


/**
 * Parse \p parmfile command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_parmfile(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;

	arg_ptr = parser_next_arg(next, &arg_len);

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(arg_ptr+arg_len) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	cfg_strncpy(&bentry->parmfile, arg_ptr, arg_len);

	return P_OK;
}


/**
 * Parse \p lock command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_lock(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(next) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	bentry->locked = 1;

	return P_OK;
}


/**
 * Parse \p pause command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_pause(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	const char *arg_ptr;
	size_t arg_len;

	// find start of title
	arg_ptr = next + strspn(next, P_WHITESPACE);

	// remove trailing whitespaces
	arg_len = strlen(arg_ptr);
	for (; arg_len > 0 && strchr(P_WHITESPACE,
		*(arg_ptr+arg_len-1)) != NULL; arg_len--);

	cfg_strncpy(&bentry->pause, arg_ptr, arg_len);

	return P_OK;
}


/**
 * Parse \p halt command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_halt(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	if (bentry->action != -1) {
		parser_set_errmsg(context, "invalid boot entry");
		return P_ERR;
	}

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(next) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	bentry->action = HALT;

	return P_OK;
}


/**
 * Parse \p reboot command and update configuration structure.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_reboot(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	if (bentry->action != -1) {
		parser_set_errmsg(context, "invalid boot entry");
		return P_ERR;
	}

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(next) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	bentry->action = REBOOT;

	return P_OK;
}


/**
 * Parse closing curly brace in boot entry section and validate boot entry
 * settings.
 *
 * \param[in,out] bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_cbrace(struct cfg_bentry *bentry, struct parser_context *context,
    const char *next)
{
	if (context->mode == PARSE_KBOOT_BOOT_ENTRY)
	{
		// in parse boot entry mode closing braces are not allowed
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	// check that there are no more arguments on this line
	if (*parser_skip_whitespace(next) != '\0') {
		parser_set_errmsg(context, "syntax error");
		return P_ERR;
	}

	// if boot action is set to -1 no boot method was specified
	if (bentry->action == -1) {
		parser_set_errmsg(context,
		    "no boot method specified for boot entry");
		return P_ERR;
	}

	// make sure title string is set
	if (!strlen(bentry->title)) {
		parser_set_errmsg(context,
		    "no title string in boot entry set");
		return P_ERR;
	}

	return P_OK;
}


/**
 * Parse commands within boot entry section and update configuration structure.
 *
 * \param[out]  bentry    Pointer to cfg_bentry structure to be configured.
 * \param[in]   context   Pointer to parser_context structure.
 * \return      \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_boot_entry(struct cfg_bentry *bentry, struct parser_context *context)
{
	enum p_return pret;
	int token;
	const char *ptr;

	// Set boot action to -1 to indicate that boot action
	// needs to be determined.
	bentry->action = -1;

	while (!(pret=parser_readln_notempty(context)) ) {
		token = parser_find_token(context->line, config_cmds, &ptr);
		if (token == -1) {
			parser_set_errmsg(context, "syntax error");
			return P_ERR;
		}
		switch (token) {
		case T_INCLUDE:
			pret = parse_cmd_include(context, ptr);
			break;
		case T_TITLE:
			pret = parse_cmd_title(bentry, context, ptr);
			break;
		case T_LABEL:
			pret = parse_cmd_label(bentry, context, ptr);
			break;
		case T_ROOT:
			pret = parse_cmd_root(bentry, context, ptr);
			break;
		case T_KERNEL:
			pret = parse_cmd_kernel(bentry, context, ptr);
			break;
		case T_INITRD:
			pret = parse_cmd_initrd(bentry, context, ptr);
			break;
		case T_CMDLINE:
			pret = parse_cmd_cmdline(bentry, context, ptr);
			break;
		case T_PARMFILE:
			pret = parse_cmd_parmfile(bentry, context, ptr);
			break;
		case T_LOCK:
			pret = parse_cmd_lock(bentry, context, ptr);
			break;
		case T_PAUSE:
			pret = parse_cmd_pause(bentry, context, ptr);
			break;
		case T_HALT:
			pret = parse_cmd_halt(bentry, context, ptr);
			break;
		case T_REBOOT:
			pret = parse_cmd_reboot(bentry, context, ptr);
			break;
		case T_CBRACE:
			return parse_cmd_cbrace(bentry, context, ptr);

		default:
			parser_set_errmsg(context, "syntax error");
			return P_ERR;
			break;
		}
		if (pret)
			break;
	}

	switch (context->mode) {
	case PARSE_KBOOT_BOOT_ENTRY:
		if (pret == P_EOF)
			return P_OK;
		else
			return pret;

	default:
		if (pret == P_EOF) {
			parser_set_errmsg(context, "unexpected end of file");
			return P_ERR;
		}
		else
			return pret;
	}
}


/**
 * Parse \p boot_entry command and update configuration structure.
 *
 * \param[in,out] config    Pointer to cfg_toplevel structure to be configured.
 * \param[in]     context   Pointer to parser_context structure.
 * \param[in]     next      Pointer to first character after command token.
 * \return        \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_cmd_boot_entry(struct cfg_toplevel *config,
    struct parser_context *context,
    const char *next)
{
	enum p_return pret;
	struct cfg_bentry bentry;

	pret = parse_next_ocbrace(context, next);
	if (pret)
		return pret;

	cfg_bentry_init(&bentry);
	pret = parse_boot_entry(&bentry, context);
	if (!pret)
		cfg_add_bentry(config, &bentry);

	cfg_bentry_destroy(&bentry);

	return pret;
}


/**
 * Parse kboot configuration file provided in parser context and update
 * configuration structure.
 *
 * \param[out] config    Pointer to cfg_toplevel structure to be configured.
 * \param[in]  context   Pointer to parser_context structure.
 * \return     \p P_OK on success, \p P_ERR on error
 */

static enum p_return
parse_global(struct cfg_toplevel *config, struct parser_context *context)
{
	enum p_return pret;
	int token;
	const char *ptr;

	while (!(pret = parser_readln_notempty(context))) {
		token = parser_find_token(context->line, config_cmds, &ptr);
		if (token == -1) {
			parser_set_errmsg(context, "syntax error");
			return P_ERR;
		}
		switch (token) {
		case T_DEFAULT:
			pret = parse_cmd_default(config, context, ptr);
			break;
		case T_TIMEOUT:
			pret = parse_cmd_timeout(config, context, ptr);
			break;
		case T_PASSWORD:
			pret = parse_cmd_password(config, context, ptr);
			break;
		case T_USERINTERFACE:
			pret = parse_cmd_userinterface(config, context, ptr);
			break;
		case T_INCLUDE:
			pret = parse_cmd_include(context, ptr);
			break;
		case T_BOOT_ENTRY:
			pret = parse_cmd_boot_entry(config, context, ptr);
			break;

		default:
			parser_set_errmsg(context, "syntax error");
			return P_ERR;
			break;
		}
		if (pret)
			break;
	}

	switch (pret) {
	case P_EOF:
		return P_OK;

	default:
		return pret;
	}
}


/**
 * Parse kboot configuration file identified by URI. On success \p NULL
 * is returned. If the configuration file contains errors the return
 * value is a dynamically allocated string with the parser error message.
 * The caller is responsible to free this memory buffer with \p free().
 *
 * \param[out] config   Pointer to unintialized cfg_toplevel structure
 *                      to return parser output.
 * \param[in]  uri      String with configuration file URI.
 * \return     \p NULL on success, in case of error dynamically allocated
 *             error message.
 */

char *
parse_kboot(struct cfg_toplevel *config, const char *uri)
{
	char *errmsg = NULL;
	int idx;
	struct parser_context context;

	cfg_init(config);
	parser_init(&context);
	context.mode = PARSE_KBOOT_CONFIGURATION;

	if (parser_open_incl_uri(&context, uri)) {
		cfg_strinit(&errmsg);
		cfg_strprintf(&errmsg, "Error accessing config file '%s': %s",
		    uri, context.errmsg);
		goto cleanup;
	}

	if (parse_global(config, &context)) {
		cfg_strinit(&errmsg);
		cfg_strprintf(&errmsg,
		    "Error parsing config URI '%s': %s in line %d.",
		    context.incl[context.depth-1].uri,
		    context.errmsg,
		    context.incl[context.depth-1].lineno);
		goto cleanup;
	}

	// resolve default boot label
	if (strspn(context.boot_default, "0123456789") ==
	    strlen(context.boot_default) &&
	    sscanf(context.boot_default, "%d", &idx) == 1) {
		if (idx<0 || idx>=config->bentry_count) {
			cfg_strinitcpy(&errmsg, "Invalid default boot index.");
			goto cleanup;
		}
		config->boot_default = idx;
	} else {
		for (idx = 0; idx<config->bentry_count; idx++) {
			if (strcmp(context.boot_default,
				config->bentry_list[idx].label) == 0) {
				config->boot_default = idx;
				break;
			}
		}
		if (idx == config->bentry_count) {
			cfg_strinitcpy(&errmsg, "Invalid default boot label.");
			goto cleanup;
		}
	}

 cleanup:
	parser_destroy(&context);

	return errmsg;
}


/**
 * Parse kboot boot entry provided in \p str. On success \p NULL
 * is returned. If the boot entry contains errors the return value
 * is a dynamically allocated string with the parser error message.
 * The caller is responsible to free this memory buffer with \p free().
 *
 * \param[out] bentry   Pointer to uninitialized cfg_bentry structure
 *                      to return parser output.
 * \param[in]  str      String with boot entry commands
 * \return     \p NULL on success, in case of error dynamically allocated
 *             error message.
 */

char *parse_kboot_boot_entry(struct cfg_bentry *bentry, const char *str)
{
	char *errmsg = NULL;
	struct parser_context context;

	cfg_bentry_init(bentry);
	parser_init(&context);
	context.mode = PARSE_KBOOT_BOOT_ENTRY;

	if (parser_open_incl_string(&context, str))
		goto cleanup;

	if (parse_boot_entry(bentry, &context)) {
		cfg_strinit(&errmsg);
		cfg_strprintf(&errmsg,
		    "Error parsing boot entry: %s in line %d.",
		    context.errmsg,
		    context.incl[context.depth-1].lineno);
		goto cleanup;
	}

	if (bentry->action == -1) {
		cfg_strinitcpy(&errmsg,
		    "no boot method specified for boot entry");
		goto cleanup;
	}

 cleanup:
	parser_destroy(&context);

	return errmsg;
}
