/*!
 * (C) Copyright IBM Corp. 2005, 2005
 *
 * \file ui_linemode.c
 * \brief kboot user interface module for line mode terminals
 *
 * $Id: ui_linemode.c,v 1.14 2006/02/13 16:08:56 loehr Exp $
 *
 * Author(s): Michael Loehr   (mloehr@de.ibm.com)
 *
 */


#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include "config.h"

char *arg0; //!< global variable with pointer to argv[0]

FILE *console_in = NULL;          /*!< for screen input */
FILE *console_out = NULL;         /*!< for screen output */


/**
 * Signal handler for SIGTERM
 */

void sigtermhandler()
{
	fprintf(console_out, "entry has been selected via timeout "
	    "or another user interface!\n");
	exit(0);
}


/**
 * Ask the user to enter a password and test against passwd
 *
 * \param[in] passwd  Password
 * \return    CFG_RETURN_OK if the user entered the correct password,
 *            CFG_RETURN_ERROR otherwise
 */

int check_password(const char *passwd)
{
	char input[CFG_STR_MAX_LEN] = ""; /* input line from user */
	char passwd_entered[CFG_STR_MAX_LEN] = ""; /* password from user */

	fprintf(console_out, "\nPlease enter the password:\n");
	fgets(input, CFG_STR_MAX_LEN, console_in);
	sscanf(input, " %s", passwd_entered);

	if (strcmp(passwd, passwd_entered) == 0) {
		fprintf(console_out, "\nPassword OK!\n");
		return CFG_RETURN_OK;
	}

	fprintf(console_out, "\nPassword incorrect!\n");
	return CFG_RETURN_ERROR;
}


/**
 * This program implements a simple linemode user interface. It collects all
 * input data that is required to display a boot selection menu from
 * environment variables and prints the selected entry to stdout.
 */

int main(int argc, char *argv[])
{
	arg0 = argv[0];                   /* make MEM_ASSERT happy */
	struct cfg_toplevel* my_toplevel = NULL; /* boot info from env */
	struct cfg_bentry* selected_bentry = NULL;
	int i = 0;                        /* multi purpose loop counter */
	int selected = -1;                /* selected boot menu entry */
	int selection_ok = CFG_RETURN_ERROR;
	char input[CFG_STR_MAX_LEN] = ""; /* input line from user */
	char *device_name = NULL;         // for user interaction
	char *message = NULL;             // message to be displayed

	cfg_strinit(&message);
	cfg_strinit(&device_name);
	my_toplevel = cfg_new();
	cfg_get_env(my_toplevel);
	cfg_get_env_str(CFG_STARTUP_MSG, &message);

	/* now we know which device to open! */
	if (argv[1] != NULL) {
		cfg_strcpy(&device_name, argv[1]);
	} else {
		cfg_strcpy(&device_name, "/dev/console");
	}

	console_in = fopen(device_name, "r");  // open the terminal keyboard
	console_out = fopen(device_name, "w"); // open the terminal screen

	if (!console_in || !console_out) {
		fprintf(stderr, "Unable to open terminal\n");
		exit(1);
	}

	signal(SIGTERM, sigtermhandler);

	do {
		selection_ok = CFG_RETURN_ERROR;
		fprintf(console_out, "\nWelcome to kboot!\n\n");
		if (strlen(message))
			fprintf(console_out, "%s\n\n", message);
		fprintf(console_out,
		    "The following boot options are available:\n\n");
		/* show the boot selection menu */
		for (i = 0; i < my_toplevel->bentry_count; i++) {
			selected_bentry = &(my_toplevel->bentry_list[i]);
			if (i == my_toplevel->boot_default )
				fprintf(console_out, "->");
			else
				fprintf(console_out, "  ");
			if (selected_bentry->locked)
				fprintf(console_out, "[");
			else
				fprintf(console_out, " ");
			fprintf(console_out, "%d %s", i+1,
			    selected_bentry->title );
			if (selected_bentry->locked)
				fprintf(console_out, "]\n");
			else
				fprintf(console_out, " \n");
		}

		fprintf(console_out, "\nPlease enter your selection:\n");
		if (my_toplevel->timeout > 0) {
			fprintf(console_out,
			    "(default will be selected in %d seconds) \n",
			    my_toplevel->timeout);
		}
		fgets(input, CFG_STR_MAX_LEN, console_in);

		if (sscanf(input, " %d", &selected) == 1)
			selected--; // internal representation is shifted by -1
		else
			selected = -1; // initialize to a safe invalid value
	 
		// check if selection and password are ok
		if (selected >= 0 && selected < my_toplevel->bentry_count) {
			selected_bentry =
				&(my_toplevel->bentry_list[selected]);
			if (selected_bentry->locked) {
				selection_ok =
					check_password(my_toplevel->password);
			} else
				selection_ok = CFG_RETURN_OK;
		} else
			fprintf(console_out, "Invalid input.\n");

	} while (selection_ok != CFG_RETURN_OK);

	/* check if pause exists and wait for user input and continue */
	if (selected_bentry->pause && strlen(selected_bentry->pause)>0) {
		fprintf(console_out, "\n%s\n", selected_bentry->pause);
		fprintf(console_out, "(press enter to continue)\n");
		fgets(input, CFG_STR_MAX_LEN, console_in);
	}

	/* now write selected entry to stdout */
	cfg_bentry_print(selected_bentry);

	fclose(console_in);
	fclose(console_out);
	cfg_strfree(&message);
	cfg_strfree(&device_name);

	exit(0);
}

